;;; "wbtab.scm" database tables using WB b-trees.
; Copyright 1996, 2000 Aubrey Jaffer
;
;Permission to copy this software, to redistribute it, and to use it
;for any purpose is granted, subject to the following restrictions and
;understandings.
;
;1.  Any copy made of this software must include this copyright notice
;in full.
;
;2.  I have made no warrantee or representation that the operation of
;this software will be error-free, and I am under no obligation to
;provide any services, by way of maintenance, update, or otherwise.
;
;3.  In conjunction with products arising from the use of this
;material, there shall be no use of my name in any advertising,
;promotional, or sales literature without prior written consent in
;each case.

;;;; *catalog* needs to be informed of the binding for 'wb-table.
;;; See "scm/mkimpcat.scm".
(require 'wb)

;;; Duplicate definitions from "wb/wbdefs.scm", which should be moved to
;;; "wb/db.c".
(define WB:NUM-SEGS 10)
(define WB:MAXERR -100)
(define (wb:err? x) (and (number? x) (negative? x) (>= x WB:MAXERR) x))

;;; WB-SEG:LOCKS has one extra location at end for loop end test
(define wb-seg:locks (make-vector (+ 1 WB:NUM-SEGS) #f))
(define wb-seg:files (make-vector (+ 1 WB:NUM-SEGS) #f))
(define wb-seg:roots (make-vector (+ 1 WB:NUM-SEGS) #f))
(define wb-seg:mut?s (make-vector (+ 1 WB:NUM-SEGS) #f))
(do ((i (+ -2 (vector-length wb-seg:locks)) (+ -1 i)))
    ((negative? i))
  (vector-set! wb-seg:locks i (make-arbiter i)))

(define wb-table
(let ((wb:make-handle list)
      (wb:handle->base-id car)
      (wb:handle->bt cadr)
      (wb:catalog-id 0)
      (wb:root-name "rwb")
      (wb:true (string (integer->char 1) #\T))
      (wb:false (string (integer->char 1) #\F))
      (wb:s31 (string (integer->char 31)))
      (wb:s30 (string (integer->char 30)))
      (wb:s1 (string (integer->char 1) #\1))
      (wb:s0 (string (integer->char 0)))
      (wb:c1 (integer->char 1))
      (wb:c0 (integer->char 0))
      (wb:free-id ""))

(define (wb:find-free-seg)
  (do ((i 0 (+ 1 i))
       (arb (vector-ref wb-seg:locks 0)
	    (vector-ref wb-seg:locks (+ 1 i))))
      ((or (not arb) (try-arbiter arb))
       (if arb i #f))))
(define (wb:release-seg seg)
  (and seg
       (release-arbiter (vector-ref wb-seg:locks seg))
       #f))

;;; Because B-trees grow with depth only very slowly, we might as well
;;; put everything into one B-tree named "rwb".

(define (wb:make-base filename dim types)
  (define segnum (wb:find-free-seg))
  (cond ((not segnum) #f)
	((wb:err? (make-seg segnum filename 2048)) (wb:release-seg segnum) #f)
	((wb:err? (open-seg segnum filename 2)) (wb:release-seg segnum) #f)
	((wb:err? (bt:put! (create-db segnum #\T wb:root-name)
			   wb:free-id
			   "1"))
	 (wb:release-seg segnum)
	 (slib:error 'wb:make-base "couldn't modify new base" filename)
	 #f)
	(else
	 (wb:seg-open-base segnum filename #t))))

(define (wb:open-base filename writable?)
  (define segnum (wb:find-free-seg))
  (cond ((wb:err? (open-seg segnum filename (if writable? 2 0)))
	 (wb:release-seg segnum) #f)
	(else (wb:seg-open-base segnum filename writable?))))

(define (wb:seg-open-base segnum filename writable?)
  (vector-set! wb-seg:files segnum filename)
  (vector-set! wb-seg:mut?s segnum writable?)
  (vector-set! wb-seg:roots segnum (open-db segnum wb:root-name))
  (cond ((wb:err? (vector-ref wb-seg:roots segnum))
	 (close-base segnum)
	 #f)
	(else segnum)))

(define (wb:write-base lldb filename)
  (define segnum lldb)
  (cond ((and filename
	      (equal? filename (vector-ref wb-seg:files segnum)))
	 (let ((status (close-seg segnum #f)))
	   (cond ((wb:err? status) #f)
		 ((wb:err? (open-seg segnum filename 2)) #f)
		 (else #t))))
	(else
	 ;;(slib:error 'wb:write-base "WB can't change database filename" filename)
	 #f)))

(define (wb:sync-base lldb)
  (define segnum lldb)
  (and segnum (wb:write-base lldb (vector-ref wb-seg:files segnum))))

(define (wb:close-base lldb)
  (define segnum lldb)
  (let ((status (close-seg segnum #f)))
    (cond ((wb:err? status)
	   (let ((status (close-seg segnum #t)))
	     (wb:release-seg segnum)
	     (not (wb:err? status))))
	  (else
	   (wb:release-seg segnum) #t))))

(define (wb:make-table lldb dim types)
  (define segnum lldb)
  (and (vector-ref wb-seg:mut?s segnum)
       (let* ((tns (bt:rem (vector-ref wb-seg:roots segnum) wb:free-id))
	      (base-id (and (string? tns) (string->number tns))))
	 (cond ((not tns)
		(slib:error 'wb:make-table 'wb:free-id "in use?")
		#f)
	       ((not base-id)
		(bt:put (vector-ref wb-seg:roots segnum) wb:free-id tns)
		(slib:error 'wb:make-table "free-id corrupted" base-id)
		#f)
	       ((not (bt:put (vector-ref wb-seg:roots segnum)
			     wb:free-id
			     (number->string (+ 1 base-id))))
		(slib:error 'wb:make-table "free-id lock broken")
		#f)
	       (else base-id)))))

(define (wb:base-id->prefix base-id)
  (define nstr (number->string base-id))
  (string-append (string #\T (integer->char (string-length nstr)))
		 nstr
		 (string (integer->char 1) #\D)))

(define (wb:incr-key prefix)
  (string-append prefix wb:s31))

;;; OPEN-TABLE allocates a new handle (in call to open-db) so each
;;; table handle will have its own last-block-used

(define (wb:open-table lldb base-id dim types)
  (define segnum lldb)
  (wb:make-handle (wb:base-id->prefix base-id)
		  (open-db segnum wb:root-name)))

(define (wb:kill-table lldb base-id dim types)
  (let* ((handle (wb:open-table lldb base-id dim types))
	 (prefix (wb:handle->base-id handle)))
    (not (wb:err? (bt:rem* (wb:handle->bt handle) prefix (wb:incr-key prefix))))))

(define (wb:make-keyifier-1 type)
  (case type
    ((string) (lambda (s) (string-append wb:s30 s)))
    ((symbol) (lambda (s) (string-append wb:s30 (symbol->string s))))
    ((integer number base-id)
     (lambda (n)
       (let ((str (number->string n)))
	 (string-append (string (integer->char
				 (min 30 (string-length str))))
			str))))
    ((boolean) (lambda (b) (if b wb:true wb:false)))
    ((atom)
     (lambda (obj)
       (cond ((number? obj)
	      (let ((str (number->string obj)))
		(string-append (string (integer->char
					(min 30 (string-length str))))
			       str)))
	     ((not obj) "#f")
	     ((eqv? #t obj) "#t")
	     (else (string-append wb:s30 (symbol->string obj))))))
    (else (slib:error 'wb:make-keyifier-1 'unsupported-type type))))

(define (wb:make-list-keyifier prinum types)
  (do ((tps (reverse types) (cdr tps)))
      ((>= prinum (length tps)) (set! types (reverse tps))))
  (if (= 1 prinum)
      (let ((proc (wb:make-keyifier-1 (car types))))
	(lambda (lst) (proc (car lst))))
      (let ((procs (map wb:make-keyifier-1 types)))
	(lambda (lst)
	  (apply string-append (map (lambda (f a) (f a)) procs lst))))))

(define (wb:string->atom str)
  (cond ((string->number str))
	((string-ci=? "#f" str) #f)
	((string-ci=? "#t" str) #t)
	(else (string->symbol str))))

(define (wb:string-> type)
  (case type
    ((string) identity)
    ((symbol) (lambda (str) (if (string-ci=? str "#f") #f
				(string->symbol str))))
    ((integer number base-id) (lambda (str) (string->number str)))
    ((boolean) (lambda (str) (not (equal? str "N"))))
    ((atom) wb:string->atom)
    ((expression) (lambda (str) (call-with-input-string str read)))
    (else (slib:error 'make-key-extractor 'unsupported-type type))))

(define (wb:atom->string obj)
  (cond ((number? obj) (number->string obj))
	((not obj) "#f")
	((eqv? #t obj) "#t")
	(else (symbol->string obj))))

(define (wb:->string type)
  (case type
    ((string) (lambda (s) (if s s "")))
    ((symbol) (lambda (sym) (symbol->string sym)))
    ((integer number base-id) (lambda (int) (if int (number->string int) "")))
    ((boolean) (lambda (b) (if b "Y" "N")))
    ((atom) wb:atom->string)
    ((expression) (lambda (obj) (call-with-output-string
				 (lambda (port) (write obj port)))))
    (else (slib:error 'make-key-extractor 'unsupported-type type))))

(define (wb:make-key-extractor primary-limit column-type-list index)
  (define proc (wb:string-> (list-ref column-type-list (+ -1 index))))
  (lambda (key)
    (define kend (string-length key))
    (let loop ((pos 0) (argind (+ -1 index)))
      (define flen (char->integer (string-ref key pos)))
      (set! pos (+ 1 pos))
      (cond ((= flen 30)
	     (do ((i pos (+ 1 i)))
		 ((or (>= i kend)
		      (<= 0 (char->integer (string-ref key i)) 30))
		  (set! flen (- i pos))))))
      (if (positive? argind)
	  (loop (+ flen pos) (+ -1 argind))
	  (proc (substring key pos (+ flen pos)))))))

(define (wb:make-key->list prinum types)
  (define procs
    (map wb:string-> (do ((tps (reverse types) (cdr tps)))
			 ((>= prinum (length tps)) (reverse tps)))))
  (lambda (key)
    (define kend (string-length key))
    (map (lambda (f v) (f v))
	 procs
	 (let loop ((pos 0) (argind prinum))
	   (define flen (char->integer (string-ref key pos)))
	   (set! pos (+ 1 pos))
	   (cond ((= flen 30)
		  (do ((i pos (+ 1 i)))
		      ((or (>= i kend)
			   (<= 0 (char->integer (string-ref key i)) 30))
		       (set! flen (- i pos))))))
	   (cond ((>= (+ flen pos) kend)
		  (list (substring key pos kend)))
		 ((positive? argind)
		  (cons (substring key pos (+ flen pos))
			(loop (+ flen pos)
			      (+ -1 argind))))
		 (else '()))))))

(define (list-of-false? lst)
  (cond ((null? lst) #t)
	((car lst) #f)
	(else (list-of-false? (cdr lst)))))

(define (wb:make-key-match? key-dimension column-types match-keys)
  (if (list-of-false? match-keys)
      (lambda (ckey) #t)
      (let ((keyploder (wb:make-key->list key-dimension column-types)))
	(lambda (ckey)
	  (define (wb:key-match? match-keys keys)
	    (cond ((null? match-keys) #t)
		  ((not (car match-keys))
		   (wb:key-match? (cdr match-keys) (cdr keys)))
		  ((equal? (car match-keys) (car keys))
		   (wb:key-match? (cdr match-keys) (cdr keys)))
		  ((not (procedure? (car match-keys)))
		   #f)
		  (((car match-keys) (car keys))
		   (wb:key-match? (cdr match-keys) (cdr keys)))
		  (else #f)))
	  (wb:key-match? match-keys (keyploder ckey))))))

(define (wb:ordered-for-each-key
	 handle operation key-dimension column-types match-keys)
  (let* ((bt (wb:handle->bt handle))
	 (prefix (wb:handle->base-id handle))
	 (pl (string-length prefix))
	 (prefix+ (wb:incr-key prefix))
	 (key-match? (wb:make-key-match? key-dimension column-types match-keys))
	 (maproc
	  (lambda (k v)
	    (let ((i (+ -1 (string-length k))))
	      (cond ((and (char=? #\1 (string-ref k i))
			  (char=? wb:c1 (string-ref k (+ -1 i))))
		     (and (key-match? (substring k pl (+ -1 i)))
			  (operation (substring k pl (+ -1 i)))))
		    ((char=? wb:c0 (string-ref k i))
		     (and (key-match? (substring k pl i))
			  (operation (substring k pl i))))))
	    #f)))
    (do ((res (bt:scan bt 0 prefix prefix+ maproc 1)
	      (bt:scan bt 0 (caddr res) prefix+ maproc 1)))
	((not (= -1 (car res)))))))

(define (wb:map-key handle operation key-dimension column-types match-keys)
  (define lst '())
  (wb:ordered-for-each-key handle
			   (lambda (k) (set! lst (cons (operation k) lst)))
			   key-dimension column-types match-keys)
  lst)

(define (wb:present? handle key)
  (let* ((kc (string-append (wb:handle->base-id handle) key))
	 (kcl (string-length kc))
	 (n (bt:next (wb:handle->bt handle) kc)))
    (and n
	 (<= (+ 1 kcl) (string-length n) (+ 2 kcl))
	 (string=? kc (substring n 0 kcl)))))

(define (wb:delete handle key)
  (let ((prefix (string-append (wb:handle->base-id handle) key)))
    (not (wb:err? (bt:rem* (wb:handle->bt handle)
			   prefix
			   (wb:incr-key prefix))))))

(define (wb:delete* handle key-dimension column-types match-keys)
  (let ((prefix (string-append (wb:handle->base-id handle) match-keys)))
    (not (wb:err? (bt:rem* (wb:handle->bt handle)
			   prefix
			   (wb:incr-key prefix))))))

(define (wb:make-getter prinum types)
  (case (- (length types) prinum)
    ((0) (lambda (handle key)
	   (and (bt:get (wb:handle->bt handle)
			(string-append (wb:handle->base-id handle) key wb:s0))
		'())))
    ((1) (let ((proc (wb:string-> (list-ref types prinum))))
	   (lambda (handle key)
	     (define val
	       (bt:get
		(wb:handle->bt handle)
		(string-append (wb:handle->base-id handle) key wb:s1)))
	     (if val (list (proc val)) #f))))
    (else (let ((procs (map wb:string-> (do ((i prinum (+ -1 i))
					     (tps types (cdr tps)))
					    ((<= i 0) (reverse tps))))))
	    (lambda (handle key)
	      (let* ((bt (wb:handle->bt handle))
		     (prefix (string-append (wb:handle->base-id handle) key))
		     (prefix+ (wb:incr-key prefix))
		     (lst '())
		     (maproc (lambda (k v) (set! lst (cons v lst)) #t)))
		(do ((res (bt:scan bt 0 prefix prefix+ maproc 1)
			  (bt:scan bt 0 (caddr res) prefix+ maproc 1)))
		    ((not (= -1 (car res)))
		     (if (zero? (cadr res)) #f
			 (do ((ps procs (cdr ps))
			      (ls lst (cdr ls))
			      (rl '() (cons ((car ps) (car ls)) rl)))
			     ((null? (cdr ls))
			      (cons ((car ps) (car ls)) rl))))))))))))

(define (wb:col-field i)
  (let ((is (number->string i)))
    (string-append (string (integer->char (string-length is))) is)))

(define (wb:make-putter prinum types)
  (case (- (length types) prinum)
    ((0) (lambda (handle ckey restcols)
	   (bt:put! (wb:handle->bt handle)
		    (string-append (wb:handle->base-id handle) ckey wb:s0)
		    "")))
    ((1) (let ((proc (wb:->string (list-ref types prinum))))
	   (lambda (handle ckey restcols)
	     (bt:put! (wb:handle->bt handle)
		      (string-append (wb:handle->base-id handle) ckey wb:s1)
		      (proc (car restcols))))))
    (else (let ((procs (map wb:->string (do ((i prinum (+ -1 i))
					     (tps types (cdr tps)))
					    ((<= i 0) tps)))))
	    (lambda (handle ckey restcols)
	      (define i 0)
	      (for-each
	       (lambda (proc val)
		 (set! i (+ 1 i))
		 (cond ((wb:err?
			 (bt:put! (wb:handle->bt handle)
				  (string-append (wb:handle->base-id handle)
						 ckey
						 (wb:col-field i))
				  (proc val)))
			(slib:error 'putter "couldn't put"
				    (string-append (wb:handle->base-id handle)
						   ckey
						   (wb:col-field i))
				    (proc val)))))
	       procs restcols))))))

(define (wb:supported-type? type)
  (case type
    ((base-id atom integer number boolean string symbol expression) #t)
    (else #f)))

(define (wb:supported-key-type? type)
  (case type
    ((atom integer number symbol string boolean) #t)
    (else #f)))

  (lambda (operation-name)
    (#+foo
     (lambda (proc)
       (if (procedure? proc)
	   (lambda args
	     (let ((ans (apply proc args)))
	       (if (procedure? ans)
		   (tracef ans operation-name)
		   ans)))
	   proc))
     identity
     (case operation-name
       ((make-base) wb:make-base)
       ((open-base) wb:open-base)
       ((write-base) wb:write-base)
       ((sync-base) wb:sync-base)
       ((close-base) wb:close-base)
       ((make-table) wb:make-table)
       ((open-table) wb:open-table)
       ((kill-table) wb:kill-table)
       ((make-keyifier-1) wb:make-keyifier-1)
       ((make-list-keyifier) wb:make-list-keyifier)
       ((make-key->list) wb:make-key->list)
       ((make-key-extractor) wb:make-key-extractor)
       ((supported-type?) wb:supported-type?)
       ((supported-key-type?) wb:supported-key-type?)
       ((present?) wb:present?)
       ((make-putter) wb:make-putter)
       ((make-getter) wb:make-getter)
       ((delete) wb:delete)
       ((delete*) wb:delete*)
       ((for-each-key) wb:ordered-for-each-key)
       ((map-key) wb:map-key)
       ((ordered-for-each-key) wb:ordered-for-each-key)
       ((catalog-id) wb:catalog-id)
       (else #f))))))
